from os import path

import pandas

"""
Summarizes both ResFinder and PointFinder database results into a single table.
"""


class AMRDetectionSummary:

    def __init__(self, files, resfinder_dataframe, pointfinder_dataframe=None):
        """
        Constructs an object for summarizing AMR detection results.
        :param files: The list of genome files we have scanned against.
        :param resfinder_dataframe: The pandas.DataFrame containing the ResFinder results.
        :param pointfinder_dataframe: The pandas.DataFrame containing the PointFinder results.
        """
        self._names = [path.splitext(path.basename(x))[0] for x in files]
        self._resfinder_dataframe = resfinder_dataframe

        if pointfinder_dataframe is not None:
            self._has_pointfinder = True
            self._pointfinder_dataframe = pointfinder_dataframe
        else:
            self._has_pointfinder = False

    def _compile_results(self, df):
        df_summary = df.sort_values(by=['Gene']).groupby(['Isolate ID']).aggregate(
            lambda x: {'Gene': "%s" % ', '.join(x['Gene'])})
        return df_summary[['Gene']]

    def _include_negatives(self, df):
        result_names_set = set(df.index.tolist())
        names_set = set(self._names)

        negative_names_set = names_set - result_names_set
        negative_entries = pandas.DataFrame([[x, 'None'] for x in negative_names_set],
                                            columns=('Isolate ID', 'Gene')).set_index('Isolate ID')
        return df.append(negative_entries)

    def create_summary(self, include_negatives=False):
        """
        Constructs a summary pandas.DataFrame for all ResFinder/PointFinder results.
        :param include_negatives: If True, include files with no ResFinder/PointFinder results.
        :return: A pandas.DataFrame summarizing the results.
        """
        df = None

        if self._has_pointfinder:
            if include_negatives:
                df = self._include_negatives(
                    self._compile_results(self._resfinder_dataframe.append(self._pointfinder_dataframe)))
            else:
                df = self._compile_results(self._resfinder_dataframe.append(self._pointfinder_dataframe))
        else:
            if include_negatives:
                df = self._include_negatives(self._compile_results(self._resfinder_dataframe))
            else:
                df = self._compile_results(self._resfinder_dataframe)

        df.rename(columns={'Gene': 'Genotype'}, inplace=True)

        return df.sort_index()



import logging
import os
import tempfile
from concurrent.futures import ThreadPoolExecutor

from Bio.Blast.Applications import NcbiblastnCommandline

logger = logging.getLogger('BlastHandler')

"""
Class for handling scheduling of BLAST jobs.
"""


class BlastHandler:

    def __init__(self, resfinder_database, threads, pointfinder_database=None):
        """
        Creates a new BlastHandler.
        :param resfinder_database: The staramr.blast.resfinder.ResfinderBlastDatabase for the particular ResFinder database.
        :param threads: The maximum number of threads to use, where one BLAST process gets assigned to one thread.
        :param pointfinder_database: The staramr.blast.pointfinder.PointfinderBlastDatabase to use for the particular PointFinder database.
        """
        self._resfinder_database = resfinder_database

        if threads is None:
            raise Exception("threads is None")

        self._threads = threads

        if (pointfinder_database == None):
            self._pointfinder_configured = False
        else:
            self._pointfinder_database = pointfinder_database
            self._pointfinder_configured = True

        self._thread_pool_executor = None
        self.reset()

    def reset(self):
        """
        Resets this BlastHandler.
        :return: None
        """
        if self._thread_pool_executor is not None:
            self._thread_pool_executor.shutdown()
        self._thread_pool_executor = ThreadPoolExecutor(max_workers=self._threads)
        self._resfinder_blast_map = {}
        self._pointfinder_blast_map = {}
        self._pointfinder_future_blasts = []
        self._resfinder_future_blasts = []
        self._temp_dirs = []

    def run_blasts(self, files):
        """
        Scans all files with BLAST against the ResFinder/PointFinder databases.
        :param files: The files to scan.
        :return: None
        """
        database_names_resfinder = self._resfinder_database.get_database_names()
        logger.debug("Resfinder Databases: " + str(database_names_resfinder))

        if self.is_pointfinder_configured():
            database_names_pointfinder = self._pointfinder_database.get_database_names()
            logger.debug("Pointfinder Databases: " + str(database_names_pointfinder))

        for file in files:
            logger.info("Scheduling blast for " + file)
            self._schedule_resfinder_blast(file, database_names_resfinder)
            if self.is_pointfinder_configured():
                self._schedule_pointfinder_blast(file, database_names_pointfinder)

    def _schedule_resfinder_blast(self, file, database_names):
        for database_name in database_names:
            database = self._resfinder_database.get_path(database_name)
            file_name = os.path.basename(file)
            dir = tempfile.TemporaryDirectory()

            # Forces temporary directories to not be cleaned up until program is finished
            self._temp_dirs.append(dir)

            blast_out = os.path.join(dir.name, file_name + ".blast.xml")
            self._resfinder_blast_map.setdefault(file_name, {})[database_name] = blast_out

            future_blast = self._thread_pool_executor.submit(self._launch_blast, file, database, blast_out)
            self._resfinder_future_blasts.append(future_blast)

    def _schedule_pointfinder_blast(self, file, database_names):
        for database_name in database_names:
            database = self._pointfinder_database.get_path(database_name)
            file_name = os.path.basename(file)
            dir = tempfile.TemporaryDirectory()

            # Forces temporary directories to not be cleaned up until this object is destroyed
            self._temp_dirs.append(dir)

            blast_out = os.path.join(dir.name, file_name + ".blast.xml")
            self._pointfinder_blast_map.setdefault(file_name, {})[database_name] = blast_out

            future_blast = self._thread_pool_executor.submit(self._launch_blast, file, database, blast_out)
            self._pointfinder_future_blasts.append(future_blast)

    def is_pointfinder_configured(self):
        """
        Whether or not PointFinder is being used.
        :return: True if PointFinder is being used, False otherwise.
        """
        return self._pointfinder_configured

    def get_resfinder_outputs(self):
        """
        Gets the ResFinder output files in the form of a dictionary which looks like:
            { 'input_file_name' => 'blast_results_file.xml' }
        :return: A dictionary mapping input file names to ResFinder BLAST output files.
        """

        # Forces any exceptions to be thrown if error with blasts
        for future_blast in self._resfinder_future_blasts:
            future_blast.result()
        return self._resfinder_blast_map

    def get_pointfinder_outputs(self):
        """
        Gets the PointFinder output files in the form of a dictionary which looks like:
            { 'input_file_name' => 'blast_results_file.xml' }
        :return: A dictionary mapping input file names to PointFinder BLAST output files.
        """
        if (self.is_pointfinder_configured()):
            # Forces any exceptions to be thrown if error with blasts
            for future_blast in self._pointfinder_future_blasts:
                future_blast.result()
            return self._pointfinder_blast_map
        else:
            raise Exception("Error, pointfinder has not been configured")

    def _launch_blast(self, query, db, output):
        blastn_command = NcbiblastnCommandline(query=query, db=db, evalue=0.001, outfmt=5, out=output)
        logger.debug(blastn_command)
        stdout, stderr = blastn_command()
        if stderr:
            raise Exception("error with [" + str(blastn_command) + "], stderr=" + stderr)

    def __del__(self):
        for dir in self._temp_dirs:
            logger.debug("Removing temporary directory " + str(dir))
            dir.cleanup()



import logging
from os import path

from staramr.databases.AMRDatabaseHandler import AMRDatabaseHandler
from staramr.databases.AMRDatabaseHandlerStripGitDir import AMRDatabaseHandlerStripGitDir

logger = logging.getLogger('AMRDatabaseHandlerFactory')

"""
A Class used to handle interactions with the ResFinder/PointFinder database files.
"""


class AMRDatabaseHandlerFactory:

    def __init__(self, database_dir, sub_dirs=False):
        """
        Builds a new AMRDatabaseHandlerFactory with the passed directory.
        :param database_dir: The directory containing the ResFinder/PointFinder databases.
        :param sub_dirs: If True, assumes we are using subdirectories to store databases
                            and searching for stripped git directories.
        """
        self._database_dir = database_dir
        self._git_database_dir = path.join(database_dir, 'update')
        self._git_strip_database_dir = path.join(database_dir, 'dist')
        self._sub_dirs = sub_dirs

    def get_database_handler(self, force_use_git=False):
        """
        Gets the appropriate database handler.
        :param force_use_git: Force use of git database handler.
        :return: The database handler.
        """
        if self._sub_dirs:
            if force_use_git or path.exists(self._git_database_dir):
                return AMRDatabaseHandler(self._git_database_dir)
            else:
                return AMRDatabaseHandlerStripGitDir(self._git_strip_database_dir)
        else:
            return AMRDatabaseHandler(self._database_dir)

    @classmethod
    def get_default_database_directory(cls):
        """
        Class method for getting the default database root directory.
        :return: The default database root directory.
        """
        return path.join(path.dirname(__file__), 'data')

    @classmethod
    def create_default_factory(cls):
        """
        Class method for getting the default database handler factory.
        :return: The default database handler factory.
        """
        return cls(cls.get_default_database_directory(), sub_dirs=True)



from staramr.blast.results.pointfinder.PointfinderHitHSP import PointfinderHitHSP
from staramr.blast.results.pointfinder.nucleotide.NucleotideMutationPosition import NucleotideMutationPosition


class PointfinderHitHSPRNA(PointfinderHitHSP):

    def __init__(self, file, blast_record, hit, hsp):
        """
        Creates a new PointfinderHitHSPRNA.
        :param file: The input file.
        :param blast_record: The Bio.Blast.Record this hit came from.
        :param hit: The particular Bio.Blast.Record.Alignment.
        :param hsp: The particular Bio.Blast.Record.HSP.
        """
        super().__init__(file, blast_record, hit, hsp)

    def _get_mutation_positions(self, start, database_frame, query_frame):
        return [NucleotideMutationPosition(i, self.hsp.sbjct, self.hsp.query, start, database_frame, query_frame) for i
                in self._get_match_positions()]



import logging

logger = logging.getLogger('BlastHits')

"""
Class for partitioning up blast hits into non-overlapping regions.
"""


class BlastHitPartitions:

    def __init__(self):
        """
        Creates a new object to store BLAST hit partitions.
        """
        self._partitions = []

    def append(self, hit):
        """
        Adds a new blast hit to the set of partitions.
        :param hit: The hit to add.
        :return: None
        """
        if hit.get_contig_start() > hit.get_contig_end():
            raise Exception(
                "Unsupported condition: contig start > contig end for hit (contig=" + hit.get_contig() + ", start=" +
                str(hit.get_contig_start()) + ", end=" + str(hit.get_contig_end()) + ")")
        if hit.get_query_frame() != 1:
            raise Exception("Unsupported condition: query frame is not 1 for hit (contig=" + hit.get_contig() +
                            ", start=" + str(hit.get_contig_start()) + ", end=" + str(
                hit.get_contig_end()) + ", query_frame=" + str(hit.get_query_frame()) + ")")

        partition = self._find_parition(hit)
        if (partition is None):
            self._partitions.append(self._create_new_parition(hit))
        else:
            self._add_hit_partition(hit, partition)

    def _add_hit_partition(self, hit, partition):
        if partition['name'] != hit.get_contig():
            raise Exception("Cannot add hit with different contig name to partition")

        if hit.get_contig_start() < partition['start']:
            partition['start'] = hit.get_contig_start()

        if hit.get_contig_end() > partition['end']:
            partition['end'] = hit.get_contig_end()

        partition['hits'].append(hit)

    def _find_parition(self, hit):
        for partition in self._partitions:
            if self._hit_in_parition(hit, partition):
                return partition
        return None

    def _hit_in_parition(self, hit, partition):
        if partition['name'] == hit.get_contig():
            start = hit.get_contig_start()
            end = hit.get_contig_end()
            if end > partition['start'] and end < partition['end']:
                return True
            elif start < partition['end'] and start > partition['start']:
                return True
            elif start <= partition['start'] and end >= partition['end']:
                return True
        return False

    def _create_new_parition(self, hit):
        return {
            'name': hit.get_contig(),
            'start': hit.get_contig_start(),
            'end': hit.get_contig_end(),
            'hits': [hit]
        }

    def get_hits_nonoverlapping_regions(self):
        """
        Gets BLAST hits divided up into separate lists for non-overlapping regions..
        :return: A list of BLAST hits divided up into non-overlapping regions.
        """
        partitions_list = []
        for partition in self._partitions:
            logger.debug("Partition " + repr(partition))
            partitions_list.append(partition['hits'])
        logger.debug("Done printing partitions")

        return partitions_list



"""
Classes for interacting with the (ResFinder/PointFinder) databases used to detect AMR genes.
"""
import argparse
import sys
from os import path, mkdir

from staramr.SubCommand import SubCommand
from staramr.Utils import get_string_with_spacing
from staramr.databases.AMRDatabaseHandlerFactory import AMRDatabaseHandlerFactory
from staramr.exceptions.CommandParseException import CommandParseException

"""
Base class for interacting with a database.
"""


class Database(SubCommand):

    def __init__(self, subparser, script_name):
        """
        Builds a SubCommand for interacting with databases.
        :param subparser: The subparser to use.  Generated from argparse.ArgumentParser.add_subparsers().
        :param script_name: The name of the script being run.
        """
        super().__init__(subparser, script_name)

    def _setup_args(self, arg_parser):
        arg_parser = self._subparser.add_parser('db', help='Download ResFinder/PointFinder databases')
        subparser = arg_parser.add_subparsers(dest='db_command',
                                              help='Subcommand for ResFinder/PointFinder databases.')
        arg_parser.add_argument('--version', action='store_true', dest='version',
                                help='Prints version information.', required=False)

        Build(subparser, self._script_name + " db")
        Update(subparser, self._script_name + " db")
        Info(subparser, self._script_name + " db")

        return arg_parser

    def run(self, args):
        super(Database, self).run(args)

        if args.db_command is None:
            self._root_arg_parser.print_help()


"""
Class for building a new database.
"""


class Build(Database):

    def __init__(self, subparser, script_name):
        """
        Creates a SubCommand for building a new database.
        :param subparser: The subparser to use.  Generated from argparse.ArgumentParser.add_subparsers().
        :param script_name: The name of the script being run.
        """
        super().__init__(subparser, script_name)

    def _setup_args(self, arg_parser):
        name = self._script_name
        default_dir = AMRDatabaseHandlerFactory.get_default_database_directory()
        epilog = ("Example:\n"
                  "\t" + name + " build\n"
                                "\t\tBuilds a new ResFinder/PointFinder database under " + default_dir + " if it does not exist\n\n" +
                  "\t" + name + " build --dir databases\n" +
                  "\t\tBuilds a new ResFinder/PointFinder database under databases/")

        arg_parser = self._subparser.add_parser('build',
                                                epilog=epilog,
                                                formatter_class=argparse.RawTextHelpFormatter,
                                                help='Downloads and builds databases in the given directory.')
        arg_parser.add_argument('--dir', action='store', dest='destination', type=str,
                                help='The directory to download the databases into [' + default_dir + '].',
                                default=default_dir, required=False)
        arg_parser.add_argument('--resfinder-commit', action='store', dest='resfinder_commit', type=str,
                                help='The specific git commit for the resfinder database [latest].', required=False)
        arg_parser.add_argument('--pointfinder-commit', action='store', dest='pointfinder_commit', type=str,
                                help='The specific git commit for the pointfinder database [latest].', required=False)
        return arg_parser

    def run(self, args):
        super(Build, self).run(args)

        if path.exists(args.destination):
            raise CommandParseException("Error, destination [" + args.destination + "] already exists",
                                        self._root_arg_parser)
        else:
            mkdir(args.destination)

        if args.destination == AMRDatabaseHandlerFactory.get_default_database_directory():
            database_handler = AMRDatabaseHandlerFactory.create_default_factory().get_database_handler()
        else:
            database_handler = AMRDatabaseHandlerFactory(args.destination).get_database_handler()
        database_handler.build(resfinder_commit=args.resfinder_commit, pointfinder_commit=args.pointfinder_commit)


"""
Class for updating an existing database.
"""


class Update(Database):

    def __init__(self, subparser, script_name):
        """
        Creates a SubCommand for updating an existing database.
        :param subparser: The subparser to use.  Generated from argparse.ArgumentParser.add_subparsers().
        :param script_name: The name of the script being run.
        """
        super().__init__(subparser, script_name)

    def _setup_args(self, arg_parser):
        default_dir = AMRDatabaseHandlerFactory.get_default_database_directory()
        name = self._script_name
        epilog = ("Example:\n"
                  "\t" + name + " update databases/\n"
                                "\t\tUpdates the ResFinder/PointFinder database under databases/\n\n" +
                  "\t" + name + " update -d\n" +
                  "\t\tUpdates the default ResFinder/PointFinder database under " + default_dir)
        arg_parser = self._subparser.add_parser('update',
                                                epilog=epilog,
                                                formatter_class=argparse.RawTextHelpFormatter,
                                                help='Updates databases in the given directories.')

        arg_parser.add_argument('-d', '--update-default', action='store_true', dest='update_default',
                                help='Updates default database directory (' + default_dir + ').', required=False)
        arg_parser.add_argument('--resfinder-commit', action='store', dest='resfinder_commit', type=str,
                                help='The specific git commit for the resfinder database [latest].', required=False)
        arg_parser.add_argument('--pointfinder-commit', action='store', dest='pointfinder_commit', type=str,
                                help='The specific git commit for the pointfinder database [latest].', required=False)
        arg_parser.add_argument('directories', nargs=argparse.REMAINDER)

        return arg_parser

    def run(self, args):
        super(Update, self).run(args)

        if len(args.directories) == 0:
            if not args.update_default:
                raise CommandParseException("Must pass at least one directory to update", self._root_arg_parser)
            else:
                database_handler = AMRDatabaseHandlerFactory.create_default_factory().get_database_handler(
                    force_use_git=True)
                database_handler.update(resfinder_commit=args.resfinder_commit,
                                        pointfinder_commit=args.pointfinder_commit)
        else:
            for directory in args.directories:
                database_handler = AMRDatabaseHandlerFactory(directory).get_database_handler()
                database_handler.update(resfinder_commit=args.resfinder_commit,
                                        pointfinder_commit=args.pointfinder_commit)


"""
Class for getting information from an existing database.
"""


class Info(Database):

    def __init__(self, subparser, script_name):
        """
        Creates a SubCommand for printing information about a database.
        :param subparser: The subparser to use.  Generated from argparse.ArgumentParser.add_subparsers().
        :param script_dir: The directory containing the main application script.
        :param script_name: The name of the script being run.
        """
        super().__init__(subparser, script_name)

    def _setup_args(self, arg_parser):
        name = self._script_name
        default_dir = AMRDatabaseHandlerFactory.get_default_database_directory()
        epilog = ("Example:\n"
                  "\t" + name + " info\n"
                                "\t\tPrints information about the default database in " + default_dir + "\n\n" +
                  "\t" + name + " info databases\n" +
                  "\t\tPrints information on the database stored in databases/")
        arg_parser = self._subparser.add_parser('info',
                                                epilog=epilog,
                                                formatter_class=argparse.RawTextHelpFormatter,
                                                help='Prints information on databases in the given directories.')
        arg_parser.add_argument('directories', nargs=argparse.REMAINDER)

        return arg_parser

    def run(self, args):
        super(Info, self).run(args)

        if len(args.directories) == 0:
            database_handler = AMRDatabaseHandlerFactory.create_default_factory().get_database_handler()
            sys.stdout.write(get_string_with_spacing(database_handler.info()))
        elif len(args.directories) == 1:
            database_handler = AMRDatabaseHandlerFactory(args.directories[0]).get_database_handler()
            sys.stdout.write(get_string_with_spacing(database_handler.info()))
        else:
            for directory in args.directories:
                database_handler = AMRDatabaseHandlerFactory(directory).get_database_handler()
                sys.stdout.write(get_string_with_spacing(database_handler.info()))



import unittest

from staramr.blast.results.pointfinder.codon.CodonMutationPosition import CodonMutationPosition


class CodonMutationPositionTest(unittest.TestCase):

    def testMutationPositionStartCodon1(self):
        mutation_position = 0
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "TTCGATCGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'TTC', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'F', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1F', 'Incorrect string')

    def testMutationPositionMiddleCodon1(self):
        mutation_position = 1
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "AGCGATCGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 2, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'AGC', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'S', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1S', 'Incorrect string')

    def testMutationPositionEndCodon1(self):
        mutation_position = 2
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "ATGGATCGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 3, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'ATG', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'M', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1M', 'Incorrect string')

    def testMutationPositionStartCodon2(self):
        mutation_position = 3
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "ATCAATCGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 4, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 2, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 2, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'GAT', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'AAT', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'D', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'N', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'D2N', 'Incorrect string')

    def testMutationPositionEndCodon2(self):
        mutation_position = 5
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "ATCGACCGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 6, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 2, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 2, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'GAT', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'GAC', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'D', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'D', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'D2D', 'Incorrect string')

    def testMutationPositionStartCodon3(self):
        mutation_position = 6
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "ATCGATGGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 7, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 3, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 3, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'CGA', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'GGA', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'R', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'G', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'R3G', 'Incorrect string')

    def testMutationPositionStartCodon1StartMethionine(self):
        mutation_position = 0
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "ATGGATCGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 1, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'ATG', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'M', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1M', 'Incorrect string')

    def testMutationPositionStartCodon1Stop(self):
        mutation_position = 2
        # @formatter:off
        database_string = "TACGATCGA"
        query_string    = "TAAGATCGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 3, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'TAC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'TAA', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'Y', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), '*', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'Y1*', 'Incorrect string')

    def testMutationPositionGapStart(self):
        mutation_position = 0
        # @formatter:off
        database_string = "ATCG"
        query_string    = "-TCG"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 1, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), '-TC', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'X', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1X', 'Incorrect string')

    def testMutationPositionGapMiddle(self):
        mutation_position = 1
        # @formatter:off
        database_string = "ATCG"
        query_string    = "A-CG"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 2, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'A-C', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'X', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1X', 'Incorrect string')

    def testMutationPositionGapEnd(self):
        mutation_position = 2
        # @formatter:off
        database_string = "ATCG"
        query_string    = "AT-G"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 3, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'AT-', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'X', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1X', 'Incorrect string')

    def testMutationPositionGapMiddleEnd(self):
        mutation_position = 2
        # @formatter:off
        database_string = "ATCGG"
        query_string    = "AT--G"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 3, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'AT-', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'X', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1X', 'Incorrect string')

    def testMutationPositionGapStartMiddleEnd(self):
        mutation_position = 3
        # @formatter:off
        database_string = "CCCATCGAC"
        query_string    = "CCC---GAC"
        #@formatter:on
        # @formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 4, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 2, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 2, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), '---', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'X', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I2X', 'Incorrect string')

    def testMutationPositionGapPreviousCodon(self):
        mutation_position = 3
        # @formatter:off
        database_string = "CCCATCGACT"
        query_string    = "CC----GACT"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 4, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 2, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 2, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), '---', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'X', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I2X', 'Incorrect string')

    def testMutationPositionGapLargerPreviousCodon(self):
        mutation_position = 3
        # @formatter:off
        database_string = "CCCATCGACTT"
        query_string    = "C-----GACTT"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 4, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 2, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 2, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), '---', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'X', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I2X', 'Incorrect string')

    def testMutationPositionGapBefore(self):
        mutation_position = 3
        # @formatter:off
        database_string = "CCCATCGAC"
        query_string    = "-CCA--GAC"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 4, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 2, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 2, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'A--', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'X', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I2X', 'Incorrect string')

    def testMutationPositionGapBeforeAfter(self):
        mutation_position = 3
        # @formatter:off
        database_string = "CCCATCGACT"
        query_string    = "-CCA--GA-T"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 4, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 2, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 2, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'A--', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'X', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I2X', 'Incorrect string')

    def testMutationPositionGapReferenceStart(self):
        mutation_position = 0
        # @formatter:off
        database_string = "-TCG"
        query_string    = "ATCG"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 1, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), '-TC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'ATC', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'X', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'I', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'X1I', 'Incorrect string')

    def testMutationPositionGapReferenceMiddle(self):
        mutation_position = 1
        # @formatter:off
        database_string = "A-CG"
        query_string    = "ATCG"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 2, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'A-C', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'ATC', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'X', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'I', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'X1I', 'Incorrect string')

    def testMutationPositionGapReferenceEnd(self):
        mutation_position = 2
        # @formatter:off
        database_string = "AT-G"
        query_string    = "ATCG"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 3, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'AT-', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'ATC', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'X', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'I', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'X1I', 'Incorrect string')

    def testMutationPositionStartCodon1DBNegative(self):
        mutation_position = 8
        # @formatter:off
        database_string = "TCGATCGAT" # rc("ATCGATCGA")
        query_string    = "TCGATCGAA"
        #@formatter:on
        database_start = 9
        database_frame = -1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 1, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'TTC', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'F', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1F', 'Incorrect string')

    def testMutationPositionMiddleCodon1DBNegative(self):
        mutation_position = 7
        # @formatter:off
        database_string = "TCGATCGAT" # rc("ATCGATCGA")
        query_string    = "TCGATCGGT"
        #@formatter:on
        database_start = 9
        database_frame = -1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 2, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'ACC', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'T', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1T', 'Incorrect string')

    def testMutationPositionEndCodon1DBNegative(self):
        mutation_position = 6
        # @formatter:off
        database_string = "TCGATCGAT" # rc("ATCGATCGA")
        query_string    = "TCGATCTAT"
        #@formatter:on
        database_start = 9
        database_frame = -1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 3, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'ATA', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'I', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1I', 'Incorrect string')

    def testMutationPositionStartCodon2DBNegative(self):
        mutation_position = 5
        # @formatter:off
        database_string = "TCGATCGAT" # rc("ATCGATCGA")
        query_string    = "TCGATTGAT"
        #@formatter:on
        database_start = 9
        database_frame = -1
        query_frame = 1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 4, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 2, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 2, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'GAT', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'AAT', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'D', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'N', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'D2N', 'Incorrect string')

    def testMutationPositionStartCodon1QNegative(self):
        mutation_position = 0
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "TTCGATCGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = -1
        mutation = CodonMutationPosition(mutation_position, database_string, query_string, database_start,
                                         database_frame, query_frame)

        self.assertEqual(mutation.get_nucleotide_position(), 1, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_codon_start(), 1, 'Incorrect codon start')
        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect mutation start')
        self.assertEqual(mutation.get_database_codon(), 'ATC', 'Incorrect database codon')
        self.assertEqual(mutation.get_query_codon(), 'TTC', 'Incorrect query codon')
        self.assertEqual(mutation.get_database_mutation(), 'I', 'Incorrect database amino acid')
        self.assertEqual(mutation.get_query_mutation(), 'F', 'Incorrect query amino acid')
        self.assertEqual(mutation.get_mutation_string_short(), 'I1F', 'Incorrect string')



"""
Utility functions that don't fit anywhere else.
"""


def get_string_with_spacing(data):
    """
    Gets a string representation of a list of key/value pairs (stored as a sub-list) with proper spacing between key/values.
    :param data: A list containing sub-lists of key/value pairs.
    :return: A string representation of the list.
    """
    string = ''
    max_width = max([len(w[0]) for w in data])

    for item in data:
        string = string + item[0].ljust(max_width) + " = " + item[1] + "\n"

    return string



#!/home/data1/tools/bin/anaconda/bin/python
import sys, os, re, math, subprocess

#Usage: ./index_pointfinder_db.py <species>

if len(sys.argv) != 2:
    sys.exit("Usage: ./index_pointfinder_db.py <species>")
else:
    species = sys.argv[1]

out_path = species 
if not os.path.exists(out_path):
    os.makedirs(out_path)

kma_path = "/home/data1/services/cgMLSTFinder/database_cgs/KMA/kma_index"
db_path = "/home/data1/services/ResFinder/database_archive/database-3.0_Test_indels/%s"%species
#out_path = /home/data1/services/ResFinder/database_archive/database-3.0_Test_indels/tuberculosis/KMA_indexes
out_path = db_path + "/" 
if not os.path.exists(out_path):
    os.makedirs(out_path)


# open gene list and write all genes into one file
with open(db_path + "/genes.txt", "r") as gene_file:
    gene_list = [gene.strip() for gene in gene_file.readlines()]

print(gene_list)

out_lst = []

for gene in gene_list:
    with open(db_path + "/" + gene + ".fsa", "r") as gene_seq:
        header = gene_seq.readline()
        if header.startswith(">") == False:
            print("ERROR, this is not a header, %s"%header)
            sys.exit(1)
        seq_lines = [seq.strip() for seq in gene_seq.readlines()]
        seq = "\n".join(seq_lines)
        out_lst.append(">" + gene + "\n" + seq)

with open("all_fasta.fsa", "w") as out_file:
    out_file.write("\n".join(out_lst))


# Index the db for kma
KMA_cmd = "%s -i %s -o %s" %(kma_path, "all_fasta.fsa", out_path + species)
os.system(KMA_cmd)




import logging
import subprocess
import time
from os import path

import git

from staramr.blast.pointfinder.PointfinderBlastDatabase import PointfinderBlastDatabase
from staramr.blast.resfinder.ResfinderBlastDatabase import ResfinderBlastDatabase

logger = logging.getLogger('AMRDatabaseHandler')

"""
A Class used to handle interactions with the ResFinder/PointFinder database files.
"""


class AMRDatabaseHandler:

    def __init__(self, database_dir):
        """
        Creates a new AMRDatabaseHandler.
        :param database_dir: The root directory for both the ResFinder/PointFinder databases.
        """
        self._database_dir = database_dir
        self._resfinder_dir = path.join(database_dir, 'resfinder')
        self._pointfinder_dir = path.join(database_dir, 'pointfinder')

        self._resfinder_url = "https://bitbucket.org/genomicepidemiology/resfinder_db.git"
        self._pointfinder_url = "https://bitbucket.org/genomicepidemiology/pointfinder_db.git"

    def build(self, resfinder_commit=None, pointfinder_commit=None):
        """
        Downloads and builds a new ResFinder/PointFinder database.
        :param resfinder_commit: The specific git commit for ResFinder.
        :param pointfinder_commit: The specific git commit for PointFinder.
        :return: None
        """
        logger.info("Cloning resfinder db [" + self._resfinder_url + "] to [" + self._resfinder_dir + "]")
        resfinder_repo = git.repo.base.Repo.clone_from(self._resfinder_url, self._resfinder_dir)

        if resfinder_commit is not None:
            logger.info("Checking out resfinder commit " + resfinder_commit)
            resfinder_repo.git.checkout(resfinder_commit)

        logger.info("Cloning pointfinder db [" + self._pointfinder_url + "] to [" + self._pointfinder_dir + "]")
        pointfinder_repo = git.repo.base.Repo.clone_from(self._pointfinder_url, self._pointfinder_dir)

        if pointfinder_commit is not None:
            logger.info("Checking out pointfinder commit " + pointfinder_commit)
            pointfinder_repo.git.checkout(pointfinder_commit)

        self._blast_format()

    def update(self, resfinder_commit=None, pointfinder_commit=None):
        """
        Updates an existing ResFinder/PointFinder database to the latest revisions (or passed specific revisions).
        :param resfinder_commit: The specific git commit for ResFinder.
        :param pointfinder_commit: The specific git commit for PointFinder.
        :return: None
        """

        if not path.exists(self._database_dir):
            self.build(resfinder_commit=resfinder_commit, pointfinder_commit=pointfinder_commit)
        else:
            resfinder_repo = git.Repo(self._resfinder_dir)
            pointfinder_repo = git.Repo(self._pointfinder_dir)

            logger.info("Updating " + self._resfinder_dir)
            resfinder_repo.heads.master.checkout()
            resfinder_repo.remotes.origin.pull()

            if resfinder_commit is not None:
                logger.info("Checking out resfinder commit " + resfinder_commit)
                resfinder_repo.git.checkout(resfinder_commit)

            resfinder_repo.git.reset('--hard')

            logger.info("Updating " + self._pointfinder_dir)
            pointfinder_repo.heads.master.checkout()
            pointfinder_repo.remotes.origin.pull()

            if pointfinder_commit is not None:
                logger.info("Checking out pointfinder commit " + pointfinder_commit)
                pointfinder_repo.git.checkout(pointfinder_commit)

            resfinder_repo.git.reset('--hard')

            self._blast_format()

    def info(self):
        """
        Gets information on the ResFinder/PointFinder databases.
        :return: Database information as a list containing key/value pairs.
        """
        data = []

        resfinder_repo = git.Repo(self._resfinder_dir)
        resfinder_repo_head = resfinder_repo.commit('HEAD')

        data.append(['resfinder_db_dir', self._resfinder_dir])
        data.append(['resfinder_db_url', self._resfinder_url])
        data.append(['resfinder_db_commit', str(resfinder_repo_head)])
        data.append(
            ['resfinder_db_date', time.strftime("%a, %d %b %Y %H:%M", time.gmtime(resfinder_repo_head.committed_date))])

        pointfinder_repo = git.Repo(self._pointfinder_dir)
        pointfinder_repo_head = pointfinder_repo.commit('HEAD')
        data.append(['pointfinder_db_dir', self._pointfinder_dir])
        data.append(['pointfinder_db_url', self._pointfinder_url])
        data.append(['pointfinder_db_commit', str(pointfinder_repo_head)])
        data.append(['pointfinder_db_date',
                     time.strftime("%a, %d %b %Y %H:%M", time.gmtime(pointfinder_repo_head.committed_date))])

        return data

    def _blast_format(self):

        logger.info("Formatting resfinder db")
        resfinder_db = ResfinderBlastDatabase(self._resfinder_dir)
        for path in resfinder_db.get_database_paths():
            self._make_blast_db(path)

        logger.info("Formatting pointfinder db")
        for organism_db in PointfinderBlastDatabase.build_databases(self._pointfinder_dir):
            for path in organism_db.get_database_paths():
                self._make_blast_db(path)

    def _make_blast_db(self, path):
        command = ['makeblastdb', '-in', path, '-dbtype', 'nucl', '-parse_seqids']
        logger.debug(' '.join(command))
        subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE).check_returncode()

    def get_resfinder_dir(self):
        """
        Gets the ResFinder database directory.
        :return: The ResFinder database directory.
        """
        return self._resfinder_dir

    def get_pointfinder_dir(self):
        """
        Gets the PointFinder database directory.
        :return: The PointFinder database directory.
        """
        return self._pointfinder_dir



import logging
import shutil
from os import path

import pandas

import staramr.Utils as Utils
from staramr.databases.AMRDatabaseHandler import AMRDatabaseHandler

logger = logging.getLogger('AMRDatabaseHandlerStripGitDir')

"""
A Class used to handle interactions with the ResFinder/PointFinder database files, stripping out the .git directory.
"""


class AMRDatabaseHandlerStripGitDir(AMRDatabaseHandler):

    def __init__(self, database_dir):
        """
        Creates a new AMRDatabaseHandlerStripGitDir.
        :param database_dir: The root directory for both the ResFinder/PointFinder databases.
        """
        super().__init__(database_dir)

        self._resfinder_dir_git = path.join(self._resfinder_dir, '.git')
        self._pointfinder_dir_git = path.join(self._pointfinder_dir, '.git')
        self._info_file = path.join(database_dir, 'info.ini')

    def build(self, resfinder_commit=None, pointfinder_commit=None):
        """
        Downloads and builds a new ResFinder/PointFinder database.
        :param resfinder_commit: The specific git commit for ResFinder.
        :param pointfinder_commit: The specific git commit for PointFinder.
        :return: None
        """
        super().build(resfinder_commit=resfinder_commit, pointfinder_commit=pointfinder_commit)

        database_info = super().info()

        # remove directories from info as they are unimportant here
        database_info_stripped = []
        for i in database_info:
            if i[0] != 'resfinder_db_dir' and i[0] != 'pointfinder_db_dir':
                database_info_stripped.append(i)

        self._write_database_info_to_file(database_info_stripped, self._info_file)

        logger.info("Removing " + self._resfinder_dir_git)
        shutil.rmtree(self._resfinder_dir_git)
        logger.info("Removing " + self._pointfinder_dir_git)
        shutil.rmtree(self._pointfinder_dir_git)

    def _write_database_info_to_file(self, database_info, file):
        file_handle = open(file, 'w')
        file_handle.write(Utils.get_string_with_spacing(database_info))
        file_handle.close()

    def _read_database_info_from_file(self, file):
        return pandas.read_csv(file, sep="=", index_col=False, header=None)

    def update(self, resfinder_commit=None, pointfinder_commit=None):
        """
        Updates an existing ResFinder/PointFinder database to the latest revisions (or passed specific revisions).
        :param resfinder_commit: The specific git commit for ResFinder.
        :param pointfinder_commit: The specific git commit for PointFinder.
        :return: None
        """
        raise Exception("Cannot update when .git directory has been removed")

    def info(self):
        """
        Gets information on the ResFinder/PointFinder databases.
        :return: Database information as a list containing key/value pairs.
        """
        data = self._read_database_info_from_file(self._info_file)
        data_matrix = data.as_matrix().tolist()
        data_matrix.insert(0, ['resfinder_db_dir', self._resfinder_dir])
        data_matrix.insert(3, ['pointfinder_db_dir', self._pointfinder_dir])

        return data_matrix



from staramr.detection.AMRDetection import AMRDetection

"""
A Class used to construct a particular staramr.detection.AMRDetection object.
"""


class AMRDetectionFactory:

    def __init__(self):
        pass

    def build(self, resfinder_database, blast_handler, pointfinder_database, include_negatives, output_dir=None):
        """
        Builds a new AMRDetection object.
        :param resfinder_database: The staramr.blast.resfinder.ResfinderBlastDatabase for the particular ResFinder database.
        :param blast_handler: The staramr.blast.BlastHandler to use for scheduling BLAST jobs.
        :param pointfinder_database: The staramr.blast.pointfinder.PointfinderBlastDatabase to use for the particular PointFinder database.
        :param include_negatives:  If True, include files lacking AMR genes in the resulting summary table.
        :param output_dir: The directory where output files are being written.
        :return: A new AMRDetection object.
        """
        return AMRDetection(resfinder_database, blast_handler, pointfinder_database, include_negatives,
                            output_dir=output_dir)



import Bio.Seq

from staramr.blast.results.pointfinder.MutationPosition import MutationPosition

"""
A Class defining a nucleotide-based mutation for PointFinder.
"""


class NucleotideMutationPosition(MutationPosition):

    def __init__(self, match_position, database_string, query_string, database_start, database_frame, query_frame):
        """
        Creates a new NucleotideMutationPosition.
        :param match_position: The particular position (0-based index) of the BLAST match string for this mutation.
        :param database_string: The BLAST database string.
        :param query_string: The BLAST query string.
        :param database_start: The start coordinates of the BLAST database hit.
        :param database_frame: The frame (strand) of the BLAST database.
        :param query_frame: The frame (strand) of the BLAST query.
        """
        super().__init__(match_position, database_start, database_frame, query_frame)

        self._database_nucleotide = database_string[match_position].upper()
        self._query_nucleotide = query_string[match_position].upper()

        if self._database_frame == -1:
            self._database_nucleotide = Bio.Seq.reverse_complement(self._database_nucleotide)
            self._query_nucleotide = Bio.Seq.reverse_complement(self._query_nucleotide)

    def get_type(self):
        return 'nucleotide'

    def get_mutation_position(self):
        return self.get_nucleotide_position()

    def get_database_mutation(self):
        return self._database_nucleotide

    def get_query_mutation(self):
        return self._query_nucleotide

    def get_mutation_string(self):
        return self.get_database_mutation() + ' -> ' + self.get_query_mutation()

    def __repr__(self):
        return "[database_start=" + str(self._database_start) + ", database_frame=" + str(
            self._database_frame) + ", query_frame=" + str(self._query_frame) + ", nucleotide_position=" \
               + str(self._nucleotide_position_database) + ", mutation_start=" + str(self.get_mutation_position()) \
               + ", mutation=" + self.get_mutation_string() + "]"



import abc

"""
A Class defining a nucleotide-based mutation for PointFinder.
"""


class MutationPosition:

    def __init__(self, match_position, database_start, database_frame, query_frame):
        """
        Creates a new MutationPosition.
        :param match_position: The particular position (0-based index) of the BLAST match string for this mutation.
        :param database_start: The start coordinates of the BLAST database hit.
        :param database_frame: The frame (strand) of the BLAST database.
        :param query_frame: The frame (strand) of the BLAST query.
        """
        __metaclass__ = abc.ABCMeta

        # TODO: I realise that frame and strand are different, and that I want to account for the strand here since I'm
        #  parsing BLASTN results <http://biopython.org/DIST/docs/api/Bio.Blast.Record.HSP-class.html>.
        #  But from all my testing with BioPython/BLAST parsing, it looks like even though the documentation refers
        #  to a "strand", the values I want to use as "strand" are accessible in BioPython under the "frame" labels.
        #  I need to look more into this.
        self._check_frames(database_frame, query_frame)

        self._database_start = database_start
        self._database_frame = database_frame
        self._query_frame = query_frame

        if database_frame == 1:
            self._nucleotide_position_database = database_start + match_position
        else:
            self._nucleotide_position_database = database_start - match_position

    def _check_frames(self, database_frame, query_frame):
        self._check_frame(database_frame)
        self._check_frame(query_frame)

    def _check_frame(self, frame):
        if frame not in [1, -1]:
            raise Exception("Error, frame=" + frame + " not in [1, -1].")

    def get_nucleotide_position(self):
        """
        Gets the nucleotide position in the BLAST database (1-based coords).
        :return: The nucleotide position.
        """
        return self._nucleotide_position_database

    def get_mutation_string_short(self):
        return self.get_database_mutation() + str(self.get_mutation_position()) + self.get_query_mutation()

    @abc.abstractmethod
    def get_type(self):
        """
        Gets the type of this mutation.
        :return: The type of this mutation.
        """
        pass

    @abc.abstractmethod
    def get_mutation_position(self):
        """
        Gets the position of this mutation.
        :return: The position of this mutation.
        """
        pass

    @abc.abstractmethod
    def get_database_mutation(self):
        """
        Gets the database characters corresponding to the mutation.
        :return: The database characters..
        """
        pass

    @abc.abstractmethod
    def get_query_mutation(self):
        """
        Gets the query characters corresponding to the mutation.
        :return: The query characters..
        """
        pass

    @abc.abstractmethod
    def get_mutation_string(self):
        """
        Gets the mutation as a string.
        :return: The mutation as a string.
        """
        pass



#!/home/data1/tools/bin/anaconda/bin/python
import sys, os, re, math, subprocess

db_path = "/home/data1/services/ResFinder/database_archive/database-3.0_Test_indels"
db_out_path = db_path + "/KMA_indel_rf/resfinder_db_index"
kma_path = "/home/data1/services/cgMLSTFinder/database_cgs/KMA/kma_index"


# Read database files into one file


all_fasta_filename = "%s/KMA_index_rf/all_fasta.fsa"%(db_path)
with open(all_fasta_filename, "w") as outfile:
    subprocess.call("cat %s/*.fsa >> %s"%(db_path, all_fasta_filename), shell=True)

# Index the db for kma
KMA_cmd = "%s -i %s -o %s "%(kma_path, all_fasta_filename, db_out_pat)
os.system(KMA_cmd)




import math

import Bio.Seq

from staramr.blast.results.pointfinder.MutationPosition import MutationPosition

"""
A Class defining a codon-based mutation for PointFinder.
"""


class CodonMutationPosition(MutationPosition):

    def __init__(self, match_position, database_string, query_string, database_start, database_frame, query_frame):
        """
        Creates a new CodonMutationPosition.
        :param match_position: The particular position (0-based index) of the BLAST match string for this mutation.
        :param database_string: The BLAST database string.
        :param query_string: The BLAST query string.
        :param database_start: The start coordinates of the BLAST database hit.
        :param database_frame: The frame (strand) of the BLAST database.
        :param query_frame: The frame (strand) of the BLAST query.
        """
        super().__init__(match_position, database_start, database_frame, query_frame)

        self._codon_start_database = math.ceil(self._nucleotide_position_database / 3)
        frame_shift = (self._nucleotide_position_database - 1) % 3

        self._database_codon = self._find_codon(database_string, match_position, database_frame, frame_shift)
        self._query_codon = self._find_codon(query_string, match_position, database_frame, frame_shift)

    def _find_codon(self, string, match_position, frame, frame_shift):
        if frame == 1:
            codon_start_index = match_position - frame_shift
            return string[codon_start_index:(codon_start_index + 3)].upper()
        else:
            codon_end_index = match_position + frame_shift
            return Bio.Seq.reverse_complement(string[(codon_end_index - 3 + 1):(codon_end_index + 1)].upper())

    def get_codon_start(self):
        """
        Gets the codon start in the BLAST database for PointFinder (1-based).
        :return: The codon start.
        """
        return self._codon_start_database

    def get_database_codon(self):
        """
        Gets the particular codon from the BLAST database.
        :return: The codon.
        """
        return self._database_codon

    def get_database_amino_acid(self):
        """
        Gets the corresponding amino acid from the BLAST database. If there is an indel, returns 'X'.
        :return: The amino acid from the BLAST database.
        """
        if '-' in self.get_database_codon():
            return 'X'
        else:
            return Bio.Seq.translate(self.get_database_codon(), table='Standard')

    def get_query_amino_acid(self):
        """
        Gets the corresponding amino acid from the BLAST query.  If there is an indel returns 'X'.
        :return: The amino acid from the BLAST query.
        """
        if '-' in self.get_query_codon():
            return 'X'
        else:
            return Bio.Seq.translate(self.get_query_codon(), table='Standard')

    def get_query_codon(self):
        """
        Gets the codon from the BLAST query.
        :return: The codon from the BLAST query.
        """
        return self._query_codon

    def get_mutation_position(self):
        return self.get_codon_start()

    def get_mutation_string(self):
        return self.get_database_codon() + ' -> ' + self.get_query_codon() + ' (' + self.get_database_amino_acid() \
               + ' -> ' + self.get_query_amino_acid() + ')'

    def get_database_mutation(self):
        return self.get_database_amino_acid().upper()

    def get_query_mutation(self):
        return self.get_query_amino_acid().upper()

    def get_type(self):
        return 'codon'

    def __repr__(self):
        return "[database_start=" + str(self._database_start) + ", database_frame=" + str(
            self._database_frame) + ", query_frame=" + str(self._query_frame) + ", nucleotide_position=" \
               + str(self._nucleotide_position_database) + ", codon_start=" + str(self._codon_start_database) \
               + ", codon=" + self._database_codon + "]"



import unittest

from staramr.blast.results.pointfinder.nucleotide.NucleotideMutationPosition import NucleotideMutationPosition


class NucleotideMutationPositionTest(unittest.TestCase):

    def testMutationPositionNucleotideStart(self):
        mutation_position = 0
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "TTCGATCGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = NucleotideMutationPosition(mutation_position, database_string, query_string, database_start,
                                              database_frame, query_frame)

        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_database_mutation(), 'A', 'Incorrect database mutation')
        self.assertEqual(mutation.get_query_mutation(), 'T', 'Incorrect query mutation')
        self.assertEqual(mutation.get_mutation_string_short(), 'A1T', 'Incorrect string')

    def testMutationPositionNucleotideMiddle(self):
        mutation_position = 4
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "ATCGCTCGA"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = NucleotideMutationPosition(mutation_position, database_string, query_string, database_start,
                                              database_frame, query_frame)

        self.assertEqual(mutation.get_mutation_position(), 5, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_database_mutation(), 'A', 'Incorrect database mutation')
        self.assertEqual(mutation.get_query_mutation(), 'C', 'Incorrect query mutation')
        self.assertEqual(mutation.get_mutation_string_short(), 'A5C', 'Incorrect string')

    def testMutationPositionNucleotideEnd(self):
        mutation_position = 8
        # @formatter:off
        database_string = "ATCGATCGA"
        query_string    = "ATCGATCGG"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = NucleotideMutationPosition(mutation_position, database_string, query_string, database_start,
                                              database_frame, query_frame)

        self.assertEqual(mutation.get_mutation_position(), 9, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_database_mutation(), 'A', 'Incorrect database mutation')
        self.assertEqual(mutation.get_query_mutation(), 'G', 'Incorrect query mutation')
        self.assertEqual(mutation.get_mutation_string_short(), 'A9G', 'Incorrect string')

    def testMutationPositionGapStart(self):
        mutation_position = 0
        # @formatter:off
        database_string = "ATCG"
        query_string    = "-TCG"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = NucleotideMutationPosition(mutation_position, database_string, query_string, database_start,
                                              database_frame, query_frame)

        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_database_mutation(), 'A', 'Incorrect database mutation')
        self.assertEqual(mutation.get_query_mutation(), '-', 'Incorrect query mutation')
        self.assertEqual(mutation.get_mutation_string_short(), 'A1-', 'Incorrect string')

    def testMutationPositionGapEnd(self):
        mutation_position = 3
        # @formatter:off
        database_string = "ATCG"
        query_string    = "ATC-"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = NucleotideMutationPosition(mutation_position, database_string, query_string, database_start,
                                              database_frame, query_frame)

        self.assertEqual(mutation.get_mutation_position(), 4, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_database_mutation(), 'G', 'Incorrect database mutation')
        self.assertEqual(mutation.get_query_mutation(), '-', 'Incorrect query mutation')
        self.assertEqual(mutation.get_mutation_string_short(), 'G4-', 'Incorrect string')

    def testMutationPositionGapReference(self):
        mutation_position = 0
        # @formatter:off
        database_string = "-TCG"
        query_string    = "ATCG"
        #@formatter:on
        database_start = 1
        database_frame = 1
        query_frame = 1
        mutation = NucleotideMutationPosition(mutation_position, database_string, query_string, database_start,
                                              database_frame, query_frame)

        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_database_mutation(), '-', 'Incorrect database mutation')
        self.assertEqual(mutation.get_query_mutation(), 'A', 'Incorrect query mutation')
        self.assertEqual(mutation.get_mutation_string_short(), '-1A', 'Incorrect string')

    def testMutationPositionStartDBNegative(self):
        mutation_position = 8
        # @formatter:off
        database_string = "TCGATCGAT" # rc("ATCGATCGA")
        query_string    = "TCGATCGAA"
        #@formatter:on
        database_start = 9
        database_frame = -1
        query_frame = 1
        mutation = NucleotideMutationPosition(mutation_position, database_string, query_string, database_start,
                                              database_frame, query_frame)

        self.assertEqual(mutation.get_mutation_position(), 1, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_database_mutation(), 'A', 'Incorrect database mutation')
        self.assertEqual(mutation.get_query_mutation(), 'T', 'Incorrect query mutation')
        self.assertEqual(mutation.get_mutation_string_short(), 'A1T', 'Incorrect string')

    def testMutationPositionMiddleDBNegative(self):
        mutation_position = 7
        # @formatter:off
        database_string = "TCGATCGAT" # rc("ATCGATCGA")
        query_string    = "TCGATCGGT"
        #@formatter:on
        database_start = 9
        database_frame = -1
        query_frame = 1
        mutation = NucleotideMutationPosition(mutation_position, database_string, query_string, database_start,
                                              database_frame, query_frame)

        self.assertEqual(mutation.get_mutation_position(), 2, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_database_mutation(), 'T', 'Incorrect database mutation')
        self.assertEqual(mutation.get_query_mutation(), 'C', 'Incorrect query mutation')
        self.assertEqual(mutation.get_mutation_string_short(), 'T2C', 'Incorrect string')

    def testMutationPositionEndDBNegative(self):
        mutation_position = 0
        # @formatter:off
        database_string = "TCGATCGAT" # rc("ATCGATCGA")
        query_string    = "CCGATCGAT"
        #@formatter:on
        database_start = 9
        database_frame = -1
        query_frame = 1
        mutation = NucleotideMutationPosition(mutation_position, database_string, query_string, database_start,
                                              database_frame, query_frame)

        self.assertEqual(mutation.get_mutation_position(), 9, 'Incorrect nucleotide position')
        self.assertEqual(mutation.get_database_mutation(), 'A', 'Incorrect database mutation')
        self.assertEqual(mutation.get_query_mutation(), 'G', 'Incorrect query mutation')
        self.assertEqual(mutation.get_mutation_string_short(), 'A9G', 'Incorrect string')



import logging

from staramr.blast.results.AMRHitHSP import AMRHitHSP
from staramr.blast.results.pointfinder.codon.CodonMutationPosition import CodonMutationPosition

logger = logging.getLogger('PointfinderHitHSP')

"""
Class storing a PointFinder BLAST hit/hsp.
"""


class PointfinderHitHSP(AMRHitHSP):

    def __init__(self, file, blast_record, hit, hsp):
        """
        Creates a new PointfinderHitHSP.
        :param file: The input file.
        :param blast_record: The Bio.Blast.Record this hit came from.
        :param hit: The particular Bio.Blast.Record.Alignment.
        :param hsp: The particular Bio.Blast.Record.HSP.
        """
        super().__init__(file, blast_record, hit, hsp)

    def get_gene(self):
        """
        Gets the particular gene name for the PointFinder hit.
        :return: The gene name.
        """
        return self.hit.hit_id

    def _get_match_positions(self):
        return [i for i, c in enumerate(self.hsp.match) if c == ' ']

    def _get_mutation_positions(self, start, database_frame, query_frame):
        return [CodonMutationPosition(i, self.hsp.sbjct, self.hsp.query, start, database_frame, query_frame) for i
                in self._get_match_positions()]

    def get_mutations(self):
        """
        Gets a list of NucleotideMutationPosition for the individual mutations.
        :return: A list of NucleotideMutationPosition.
        """
        start = self.hsp.sbjct_start
        database_frame = self.get_database_frame()
        query_frame = self.get_query_frame()
        return self._get_mutation_positions(start, database_frame, query_frame)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        else:
            return self.__dict__ == other.__dict__



from os import path

import pandas

from staramr.blast.results.BlastResultsParser import BlastResultsParser
from staramr.blast.results.resfinder.ResfinderHitHSP import ResfinderHitHSP

"""
Class used to parse out BLAST results for ResFinder.
"""


class BlastResultsParserResfinder(BlastResultsParser):

    def __init__(self, file_blast_map, blast_database, pid_threshold, plength_threshold, report_all=False,
                 output_dir=None):
        """
        Creates a new BlastResultsParserResfinder.
        :param file_blast_map: A map/dictionary linking input files to BLAST results files.
        :param blast_database: The particular staramr.blast.AbstractBlastDatabase to use.
        :param pid_threshold: A percent identity threshold for BLAST results.
        :param plength_threshold: A percent length threshold for results.
        :param report_all: Whether or not to report all blast hits.
        :param output_dir: The directory where output files are being written.
        """
        super().__init__(file_blast_map, blast_database, pid_threshold, plength_threshold, report_all,
                         output_dir=output_dir)

    def _create_hit(self, file, database_name, blast_record, alignment, hsp):
        return ResfinderHitHSP(file, blast_record, alignment, hsp)

    def _create_data_frame(self, results):
        df = pandas.DataFrame(results, columns=('Isolate ID', 'Gene', '%Identity', '%Overlap',
                                                'HSP Length/Total Length', 'Contig', 'Start', 'End', 'Accession'))
        return df.set_index('Isolate ID')

    def _append_results_to(self, hit, database_name, results, seq_records):
        self._append_seqrecords_to(hit, seq_records)
        results.append([hit.get_isolate_id(),
                        hit.get_gene(),
                        hit.get_pid(),
                        hit.get_plength(),
                        str(hit.get_hsp_alignment_length()) + "/" + str(hit.get_alignment_length()),
                        hit.get_contig(),
                        hit.get_contig_start(),
                        hit.get_contig_end(),
                        hit.get_accession()
                        ])

    def _get_out_file_name(self, in_file):
        if self._output_dir:
            return path.join(self._output_dir, 'resfinder_' + path.basename(in_file))
        else:
            raise Exception("output_dir is None")



"""
An Exception to be raised during command-line parsing.
"""


class CommandParseException(Exception):

    def __init__(self, msg, parser):
        """
        Constructs a new CommandParseException
        :param msg: The Exception message.
        :param parser: The argparse.ArgumentParser for the particular subparser.
        """
        super().__init__(msg)
        self._parser = parser

    def get_parser(self):
        """
        Gets the argparse.ArgumentParser for the particular subparser that threw the Exception.
        :return:
        """
        return self._parser



import argparse
import datetime
import logging
import multiprocessing
import sys
from os import path, mkdir

import pandas

from staramr.SubCommand import SubCommand
from staramr.Utils import get_string_with_spacing
from staramr.blast.BlastHandler import BlastHandler
from staramr.blast.pointfinder.PointfinderBlastDatabase import PointfinderBlastDatabase
from staramr.blast.resfinder.ResfinderBlastDatabase import ResfinderBlastDatabase
from staramr.databases.AMRDatabaseHandlerFactory import AMRDatabaseHandlerFactory
from staramr.exceptions.CommandParseException import CommandParseException

logger = logging.getLogger("Search")

"""
Class for searching for AMR resistance genes.
"""


class Search(SubCommand):
    blank = '-'

    def __init__(self, amr_detection_factory, subparser, script_name, version):
        """
        Creates a new Search sub-command instance.
        :param amr_detection_factory: A factory of type staramr.detection.AMRDetectionFactory for building necessary objects for AMR detection.
        :param subparser: The subparser to use.  Generated from argparse.ArgumentParser.add_subparsers().
        :param script_name: The name of the script being run.
        :param version: The version of this software.
        """
        super().__init__(subparser, script_name)
        self._amr_detection_factory = amr_detection_factory
        self._version = version

    def _setup_args(self, arg_parser):
        name = self._script_name
        epilog = ("Example:\n"
                  "\t" + name + " search --output-dir out *.fasta\n"
                                "\t\tSearches the files *.fasta for AMR genes using only the ResFinder database, storing results in the out/ directory.\n\n" +
                  "\t" + name + " search --pointfinder-organism salmonella --output-dir out *.fasta\n" +
                  "\t\tSearches *.fasta for AMR genes using ResFinder and PointFinder database with the passed organism, storing results in out/.")

        arg_parser = self._subparser.add_parser('search',
                                                epilog=epilog,
                                                formatter_class=argparse.RawTextHelpFormatter,
                                                help='Search for AMR genes')

        self._default_database_dir = AMRDatabaseHandlerFactory.get_default_database_directory()
        cpu_count = multiprocessing.cpu_count()

        arg_parser.add_argument('-n', '--nprocs', action='store', dest='nprocs', type=int,
                                help='The number of processing cores to use [' + str(cpu_count) + '].',
                                default=cpu_count, required=False)
        arg_parser.add_argument('--pid-threshold', action='store', dest='pid_threshold', type=float,
                                help='The percent identity threshold [98.0].', default=98.0, required=False)
        arg_parser.add_argument('--percent-length-overlap-resfinder', action='store',
                                dest='plength_threshold_resfinder', type=float,
                                help='The percent length overlap for resfinder results [60.0].', default=60.0,
                                required=False)
        arg_parser.add_argument('--percent-length-overlap-pointfinder', action='store',
                                dest='plength_threshold_pointfinder', type=float,
                                help='The percent length overlap for pointfinder results [95.0].', default=95.0,
                                required=False)
        arg_parser.add_argument('--pointfinder-organism', action='store', dest='pointfinder_organism', type=str,
                                help='The organism to use for pointfinder {' + ', '.join(
                                    PointfinderBlastDatabase.get_available_organisms()) + '} [None].', default=None,
                                required=False)
        arg_parser.add_argument('--include-negatives', action='store_true', dest='include_negatives',
                                help='Inclue negative results (those sensitive to antimicrobials) [False].',
                                required=False)
        arg_parser.add_argument('--report-all-blast', action='store_true', dest='report_all_blast',
                                help='Report all blast hits (vs. only top blast hits) [False].',
                                required=False)
        arg_parser.add_argument('-d', '--database', action='store', dest='database', type=str,
                                help='The directory containing the resfinder/pointfinder databases [' + self._default_database_dir + '].',
                                default=self._default_database_dir, required=False)
        arg_parser.add_argument('-o', '--output-dir', action='store', dest='output_dir', type=str,
                                help="The output directory for results.  If unset prints all results to stdout.",
                                default=None, required=False)
        arg_parser.add_argument('--version', action='store_true', dest='version',
                                help='Prints version information.', required=False)
        arg_parser.add_argument('files', nargs=argparse.REMAINDER)

        return arg_parser

    def _print_dataframes_to_excel(self, outfile_path, summary_dataframe, resfinder_dataframe, pointfinder_dataframe,
                                   settings_dataframe):
        writer = pandas.ExcelWriter(outfile_path, engine='xlsxwriter')

        summary_dataframe.to_excel(writer, 'Summary', freeze_panes=[1, 1], na_rep=self.blank)
        resfinder_dataframe.to_excel(writer, 'ResFinder', freeze_panes=[1, 1], na_rep=self.blank)
        if pointfinder_dataframe is not None:
            pointfinder_dataframe.to_excel(writer, 'PointFinder', freeze_panes=[1, 1], na_rep=self.blank)
        settings_dataframe.to_excel(writer, 'Settings')

        writer.save()

    def _print_dataframe_to_text_file(self, dataframe, file=None):
        file_handle = sys.stdout

        if dataframe is not None:
            if file:
                file_handle = open(file, 'w')

            dataframe.to_csv(file_handle, sep="\t", float_format="%0.2f", na_rep=self.blank)

            if file:
                file_handle.close()

    def _print_settings_to_file(self, settings, file):
        file_handle = open(file, 'w')
        file_handle.write(get_string_with_spacing(settings))
        file_handle.close()

    def run(self, args):
        super(Search, self).run(args)

        start_time = datetime.datetime.now()

        if (len(args.files) == 0):
            raise CommandParseException("Must pass a fasta file to process", self._root_arg_parser)

        hits_output_dir = None
        if args.output_dir:
            if path.exists(args.output_dir):
                raise CommandParseException("Output directory [" + args.output_dir + "] already exists",
                                            self._root_arg_parser)
            else:
                hits_output_dir = path.join(args.output_dir, 'hits')
                mkdir(args.output_dir)
                mkdir(hits_output_dir)

        if not path.isdir(args.database):
            raise CommandParseException("Database directory [" + args.database + "] does not exist")

        if args.database == AMRDatabaseHandlerFactory.get_default_database_directory():
            database_handler = AMRDatabaseHandlerFactory.create_default_factory().get_database_handler()
        else:
            database_handler = AMRDatabaseHandlerFactory(args.database)

        resfinder_database_dir = database_handler.get_resfinder_dir()
        pointfinder_database_dir = database_handler.get_pointfinder_dir()

        resfinder_database = ResfinderBlastDatabase(resfinder_database_dir)
        if (args.pointfinder_organism):
            if args.pointfinder_organism not in PointfinderBlastDatabase.get_available_organisms():
                raise CommandParseException("The only Pointfinder organism(s) currently supported are " + str(
                    PointfinderBlastDatabase.get_available_organisms()), self._root_arg_parser)
            pointfinder_database = PointfinderBlastDatabase(pointfinder_database_dir,
                                                            args.pointfinder_organism)
        else:
            pointfinder_database = None
        blast_handler = BlastHandler(resfinder_database, args.nprocs, pointfinder_database)

        amr_detection = self._amr_detection_factory.build(resfinder_database, blast_handler, pointfinder_database,
                                                          args.include_negatives, output_dir=hits_output_dir)
        amr_detection.run_amr_detection(args.files, args.pid_threshold, args.plength_threshold_resfinder,
                                        args.plength_threshold_pointfinder, args.report_all_blast)

        end_time = datetime.datetime.now()
        time_difference = end_time - start_time
        time_difference_minutes = "%0.2f" % (time_difference.total_seconds() / 60)

        logger.info("Finished. Took " + str(time_difference_minutes) + " minutes.")

        if args.output_dir:
            self._print_dataframe_to_text_file(amr_detection.get_resfinder_results(),
                                               path.join(args.output_dir, "resfinder.tsv"))
            self._print_dataframe_to_text_file(amr_detection.get_pointfinder_results(),
                                               path.join(args.output_dir, "pointfinder.tsv"))
            self._print_dataframe_to_text_file(amr_detection.get_summary_results(),
                                               path.join(args.output_dir, "summary.tsv"))

            settings = database_handler.info()
            settings.insert(0, ['command_line', ' '.join(sys.argv)])
            settings.insert(1, ['version', self._version])
            settings.insert(2, ['start_time', start_time.strftime("%Y-%m-%d %H:%M:%S")])
            settings.insert(3, ['end_time', end_time.strftime("%Y-%m-%d %H:%M:%S")])
            settings.insert(4, ['total_minutes', time_difference_minutes])
            self._print_settings_to_file(settings, path.join(args.output_dir, "settings.txt"))

            settings_dataframe = pandas.DataFrame(settings, columns=('Key', 'Value')).set_index('Key')

            self._print_dataframes_to_excel(path.join(args.output_dir, 'results.xlsx'),
                                            amr_detection.get_summary_results(), amr_detection.get_resfinder_results(),
                                            amr_detection.get_pointfinder_results(),
                                            settings_dataframe)

            logger.info("Output files in " + args.output_dir)
        else:
            self._print_dataframe_to_text_file(amr_detection.get_summary_results())



import logging
from os import path

import pandas

from staramr.blast.results.BlastResultsParser import BlastResultsParser
from staramr.blast.results.pointfinder.PointfinderHitHSP import PointfinderHitHSP
from staramr.blast.results.pointfinder.nucleotide.PointfinderHitHSPRNA import PointfinderHitHSPRNA

logger = logging.getLogger('BlastResultsParserPointfinder')

"""
A Class for parsing BLAST results specific to PointFinder.
"""
logger = logging.getLogger('BlastResultsParserPointfinder')


class BlastResultsParserPointfinder(BlastResultsParser):

    def __init__(self, file_blast_map, blast_database, pid_threshold, plength_threshold, report_all=False,
                 output_dir=None):
        """
        Creates a new BlastResultsParserPointfinder.
        :param file_blast_map: A map/dictionary linking input files to BLAST results files.
        :param blast_database: The particular staramr.blast.AbstractBlastDatabase to use.
        :param pid_threshold: A percent identity threshold for BLAST results.
        :param plength_threshold: A percent length threshold for results.
        :param report_all: Whether or not to report all blast hits.
        :param output_dir: The directory where output files are being written.
        """
        super().__init__(file_blast_map, blast_database, pid_threshold, plength_threshold, report_all,
                         output_dir=output_dir)

    def _create_hit(self, file, database_name, blast_record, alignment, hsp):
        logger.debug("database_name=" + database_name)
        if database_name == '16S_rrsD':
            return PointfinderHitHSPRNA(file, blast_record, alignment, hsp)
        else:
            return PointfinderHitHSP(file, blast_record, alignment, hsp)

    def _create_data_frame(self, results):
        df = pandas.DataFrame(results,
                              columns=('Isolate ID', 'Gene', 'Type', 'Position', 'Mutation',
                                       '%Identity', '%Overlap', 'HSP Length/Total Length', 'Contig', 'Start', 'End'))
        return df.set_index('Isolate ID')

    def _do_append(self, hit, db_mutation, results):
        results.append([hit.get_isolate_id(),
                        hit.get_hit_id() + " (" + db_mutation.get_mutation_string_short() + ")",
                        db_mutation.get_type(),
                        db_mutation.get_mutation_position(),
                        db_mutation.get_mutation_string(),
                        hit.get_pid(),
                        hit.get_plength(),
                        str(hit.get_hsp_alignment_length()) + "/" + str(hit.get_alignment_length()),
                        hit.get_contig(),
                        hit.get_contig_start(),
                        hit.get_contig_end()
                        ])

    def _append_results_to(self, hit, database_name, results, seq_records):
        database_mutations = hit.get_mutations()

        gene = hit.get_gene()

        logger.debug("")
        logger.debug("gene=" + str(gene))
        logger.debug("sbjct_start=" + str(hit.hsp.sbjct_start))
        logger.debug("sbjct_end=" + str(hit.hsp.sbjct_end))
        logger.debug("sbjct_frame=" + str(hit.get_database_frame()))
        logger.debug("query_frame=" + str(hit.get_query_frame()))
        logger.debug("query_start=" + str(hit.hsp.query_start))
        logger.debug("query_end=" + str(hit.hsp.query_end))
        for x in database_mutations:
            logger.debug("database_mutations: position=" + str(
                x.get_mutation_position()) + ", mutation=" + x.get_mutation_string())

        if database_name == '16S_rrsD':
            database_resistance_mutations = self._blast_database.get_resistance_nucleotides(gene, database_mutations)
        else:
            database_resistance_mutations = self._blast_database.get_resistance_codons(gene, database_mutations)
        logger.debug("database_resistance_mutations=" + str(database_resistance_mutations))

        logger.debug("gaps=" + str(hit.hsp.gaps))

        if len(database_resistance_mutations) == 0:
            logger.debug("No mutations for [id=" + hit.get_hit_id() + ", file=" + hit.get_file() + "]")
        else:
            for db_mutation in database_resistance_mutations:
                logger.debug("multiple resistance mutations for [" + hit.get_hit_id() + "], mutations " + str(
                    database_resistance_mutations) + ", file=" + hit.get_file() + "]")
                self._append_seqrecords_to(hit, seq_records)
                self._do_append(hit, db_mutation, results)

    def _get_out_file_name(self, in_file):
        if self._output_dir:
            return path.join(self._output_dir, 'pointfinder_' + path.basename(in_file))
        else:
            raise Exception("output_dir is None")



import os
import re

import Bio.Seq

"""
Class used to store/parse AMR BLAST hits/hsps.
"""


class AMRHitHSP:

    def __init__(self, file, blast_record, hit, hsp):
        """
        Creates a new AMRHitHSP.
        :param file: The particular file this BLAST hit came from.
        :param blast_record: The Bio.Blast.Record this hit came from.
        :param hit: The particular Bio.Blast.Record.Alignment.
        :param hsp: The particular Bio.Blast.Record.HSP.
        """
        self._file = file
        self._blast_record = blast_record
        self.hit = hit
        self.hsp = hsp

    def get_alignment_length(self):
        """
        Gets the BLAST alignment length.
        :return: The BLAST alignment length.
        """
        return self.hit.length

    def get_hsp_alignment_length(self):
        """
        Gets the BLAST HSP length.
        :return: The BLAST HSP length.
        """
        return self.hsp.align_length

    def get_pid(self):
        """
        Gets the percent identity of the HSP.
        :return: The HSP percent identity.
        """
        return (self.hsp.identities / self.hsp.align_length) * 100

    def get_plength(self):
        """
        Gets the percent length of the HSP.
        :return: The percent length of the HSP.
        """
        return (self.get_hsp_alignment_length() / self.get_alignment_length()) * 100

    def get_hit_id(self):
        """
        Gets the hit id.
        :return: The hit id.
        """
        return self.hit.hit_id

    def get_file(self):
        """
        Gets the corresponding input file.
        :return: The corresponding input file.
        """
        return self._file

    def get_isolate_id(self):
        """
        Gets isolate id for the file.
        :return: The isolate id for the file
        """
        return os.path.splitext(self._file)[0]

    def get_contig(self):
        """
        Gets the particular contig id this HSP came from in the input file.
        :return: The contig id.
        """
        re_search = re.search(r'^(\S+)', self._blast_record.query)
        return re_search.group(1)

    def get_contig_start(self):
        """
        Gets the start of the HSP in the contig in the input file.
        :return: The start of the HSP.
        """
        return self.hsp.query_start

    def get_contig_end(self):
        """
        Gets the end of the HSP in the contig from the input file.
        :return: The end of the HSP.
        """
        return self.hsp.query_end

    def get_resistance_gene_start(self):
        """
        Gets the start of the hsp to the resistance gene.
        :return: The start of the resistance gene hsp.
        """
        return self.hsp.sbjct_start

    def get_resistance_gene_end(self):
        """
        Gets the end of the hsp to the resistance gene.
        :return: The end of the resistance gene hsp.
        """
        return self.hsp.sbjct_end

    def _get_hsp_frame(self, index):
        frame = self.hsp.frame[index]
        if frame not in [1, -1]:
            raise Exception("frame=" + str(frame) + ", is unexpected")
        else:
            return frame

    def get_hsp_query(self):
        """
        Gets the query sequence from the HSP.
        :return: The query sequence (as a string) from the HSP.
        """
        return self.hsp.query

    def get_hsp_query_proper(self):
        """
        Gets the query sequence from the HSP (proper frame/strand).
        :return: The query sequence (as a string) from the HSP.
        """
        seq = self.hsp.query

        if self.get_database_frame() == -1 or self.get_query_frame() == -1:
            return Bio.Seq.reverse_complement(seq)
        else:
            return seq

    def get_database_frame(self):
        """
        Gets the frame (strand) of the BLAST database for the hit/hsp.
        :return: The frame (strand) of the BLAST database.
        """
        return self._get_hsp_frame(1)

    def get_query_frame(self):
        """
        Gets the frame (strand) of the BLAST query for the hit/hsp.
        :return: The frame (strand) of the BLAST query.
        """
        return self._get_hsp_frame(0)
